@@ -2,7 +2,7 @@ package SQL::Abstract::More;
 use strict;
 use warnings;
 
-use SQL::Abstract 1.73;
+use SQL::Abstract 1.74;
 use parent 'SQL::Abstract';
 use MRO::Compat;
 use mro 'c3'; # implements next::method
@@ -12,7 +12,11 @@ use Params::Validate  qw/validate SCALAR SCALARREF CODEREF ARRAYREF HASHREF
 use Scalar::Util      qw/blessed reftype/;
 use Carp;
 
-our $VERSION = '1.25';
+our $VERSION = '1.29';
+
+# import the "puke" function from SQL::Abstract (kind of "die")
+BEGIN {*puke = \&SQL::Abstract::puke;}
+
 
 #----------------------------------------------------------------------
 # utility function : cheap version of Scalar::Does (too heavy to be included)
@@ -22,13 +26,14 @@ my %meth_for = (
   HASH  => '%{}',
  );
 
-sub does {
+sub does ($$) {
   my ($data, $type) = @_;
   my $reft = reftype $data;
   return defined $reft && $reft eq $type
       || blessed $data && overload::Method($data, $meth_for{$type});
 }
 
+
 #----------------------------------------------------------------------
 # remove all previously defined functions
 #----------------------------------------------------------------------
@@ -74,35 +79,40 @@ my %limit_offset_dialects = (
 my %common_join_syntax = (
   '<=>' => '%s INNER JOIN %s ON %s',
    '=>' => '%s LEFT OUTER JOIN %s ON %s',
-  '<='  => '%s RIGHT JOIN %s ON %s',
+  '<='  => '%s RIGHT OUTER JOIN %s ON %s',
   '=='  => '%s NATURAL JOIN %s',
+  '>=<' => '%s FULL OUTER JOIN %s ON %s',
 );
 my %right_assoc_join_syntax = %common_join_syntax;
 s/JOIN %s/JOIN (%s)/ foreach values %right_assoc_join_syntax;
 
 # specification of parameters accepted by the new() method
 my %params_for_new = (
-  table_alias      => {type => SCALAR|CODEREF, default  => '%s AS %s'},
-  column_alias     => {type => SCALAR|CODEREF, default  => '%s AS %s'},
-  limit_offset     => {type => SCALAR|CODEREF, default  => 'LimitOffset'},
-  join_syntax      => {type => HASHREF,        default  =>
-                                                    \%common_join_syntax},
-  join_assoc_right => {type => BOOLEAN,        default  => 0},
-  max_members_IN   => {type => SCALAR,         optional => 1},
-  sql_dialect      => {type => SCALAR,         optional => 1},
+  table_alias          => {type => SCALAR|CODEREF,   default  => '%s AS %s'},
+  column_alias         => {type => SCALAR|CODEREF,   default  => '%s AS %s'},
+  limit_offset         => {type => SCALAR|CODEREF,   default  => 'LimitOffset'},
+  join_syntax          => {type => HASHREF,          default  =>
+                                                        \%common_join_syntax},
+  join_assoc_right     => {type => BOOLEAN,          default  => 0},
+  max_members_IN       => {type => SCALAR,           optional => 1},
+  multicols_sep        => {type => SCALAR|SCALARREF, optional => 1},
+  has_multicols_in_SQL => {type => BOOLEAN,          optional => 1},
+  sql_dialect          => {type => SCALAR,           optional => 1},
+  select_implicitly_for=> {type => SCALAR|UNDEF,     optional => 1},
 );
 
 # builtin collection of parameters, for various databases
 my %sql_dialects = (
- MsAccess  => { join_assoc_right => 1,
-                join_syntax      => \%right_assoc_join_syntax},
- BasisJDBC => { column_alias     => "%s %s",
-                max_members_IN   => 255                      },
- MySQL_old => { limit_offset     => "LimitXY"                },
- Oracle    => { limit_offset     => "RowNum",
-                max_members_IN   => 999,
-                table_alias      => '%s %s',
-                column_alias     => '%s %s',                 },
+ MsAccess  => { join_assoc_right     => 1,
+                join_syntax          => \%right_assoc_join_syntax},
+ BasisJDBC => { column_alias         => "%s %s",
+                max_members_IN       => 255                      },
+ MySQL_old => { limit_offset         => "LimitXY"                },
+ Oracle    => { limit_offset         => "RowNum",
+                max_members_IN       => 999,
+                table_alias          => '%s %s',
+                column_alias         => '%s %s',
+                has_multicols_in_SQL => 1,                       },
 );
 
 
@@ -134,11 +144,12 @@ my %params_for_insert = (
   -returning    => {type => SCALAR|ARRAYREF|HASHREF, optional => 1},
 );
 my %params_for_update = (
-  -table        => {type => SCALAR},
+  -table        => {type => SCALAR|SCALARREF|ARRAYREF},
   -set          => {type => HASHREF},
   -where        => {type => SCALAR|ARRAYREF|HASHREF, optional => 1},
   -order_by     => {type => SCALAR|ARRAYREF|HASHREF, optional => 1},
   -limit        => {type => SCALAR,                  optional => 1},
+  -returning    => {type => SCALAR|ARRAYREF|HASHREF, optional => 1},
 );
 my %params_for_delete = (
   -from         => {type => SCALAR},
@@ -170,10 +181,14 @@ sub new {
     $more_params{$_} ||= $dialect_params->{$_} foreach keys %$dialect_params;
   }
 
-  # check parameters
+  # check parameters for this class
   my @more_params = %more_params;
   my $more_self   = validate(@more_params, \%params_for_new);
 
+  # check some of the params for parent -- because SQLA doesn't do it :-(
+  !$params{quote_char} || exists $params{name_sep}
+    or carp "when 'quote_char' is present, 'name_sep' should be present too";
+
   # call parent constructor
   my $self = $class->next::method(%params);
 
@@ -181,8 +196,8 @@ sub new {
   $self->{$_} = $more_self->{$_} foreach keys %$more_self;
 
   # arguments supplied as scalars are transformed into coderefs
-  ref $self->{column_alias} or $self->_make_AS_through_sprintf('column_alias');
-  ref $self->{table_alias}  or $self->_make_AS_through_sprintf('table_alias');
+  ref $self->{column_alias} or $self->_make_sub_column_alias;
+  ref $self->{table_alias}  or $self->_make_sub_table_alias;
   ref $self->{limit_offset} or $self->_choose_LIMIT_OFFSET_dialect;
 
   # regex for parsing join specifications
@@ -201,6 +216,7 @@ sub new {
   return $self;
 }
 
+
 #----------------------------------------------------------------------
 # the select method
 #----------------------------------------------------------------------
@@ -211,17 +227,14 @@ sub select {
   # if got positional args, this is not our job, just delegate to the parent
   return $self->next::method(@_) if !&_called_with_named_args;
 
-  # declare variables and parse arguments;
-  my ($join_info, %aliased_columns);
+  my %aliased_columns;
+
+  # parse arguments
   my %args = validate(@_, \%params_for_select);
 
   # compute join info if the datasource is a join
-  if (ref $args{-from} eq 'ARRAY' && $args{-from}[0] eq '-join') {
-    my @join_args = @{$args{-from}};
-    shift @join_args;           # drop initial '-join'
-    $join_info   = $self->join(@join_args);
-    $args{-from} = \($join_info->{sql});
-  }
+  my $join_info = $self->_compute_join_info($args{-from});
+  $args{-from}  = \($join_info->{sql}) if $join_info;
 
   # reorganize columns; initial members starting with "-" are extracted
   # into a separate list @post_select, later re-injected into the SQL
@@ -230,9 +243,11 @@ sub select {
   push @post_select, shift @cols while @cols && $cols[0] =~ s/^-//;
   foreach my $col (@cols) {
     # extract alias, if any
-    if ($col =~ /^(.*[^|\s])  # any non-empty string, not ending with ' ' or '|'
+    if ($col =~ /^\s*         # ignore insignificant leading spaces
+                 (.*[^|\s])   # any non-empty string, not ending with ' ' or '|'
                  \|           # followed by a literal '|'
                  (\w+)        # followed by a word (the alias))
+                 \s*          # ignore insignificant trailing spaces
                  $/x) {
       $aliased_columns{$2} = $1;
       $col = $self->column_alias($1, $2);
@@ -287,24 +302,9 @@ sub select {
   # add ORDER BY if needed
   if (my $order = $args{-order_by}) {
 
-    # force scalar into an arrayref
-    $order = [$order] if not ref $order;
-
-    # restructure array data
-    if (ref $order eq 'ARRAY') {
-      my @clone = @$order;      # because we will modify items 
-
-      # '-' and '+' prefixes are translated into {-desc/asc => } hashrefs
-      foreach my $item (@clone) {
-        next if !$item or ref $item;
-        $item =~ s/^-//  and $item = {-desc => $item} and next;
-        $item =~ s/^\+// and $item = {-asc  => $item};
-      }
-      $order = \@clone;
-    }
-
-    my $sql_order = $self->where(undef, $order);
+    my ($sql_order, @orderby_bind) = $self->_order_by($order);
     $sql .= $sql_order;
+    push @bind, @orderby_bind;
   }
 
   # add LIMIT/OFFSET if needed
@@ -316,8 +316,9 @@ sub select {
     push @bind, @limit_bind;
   }
 
-  # add FOR if needed
-  $sql .= " FOR $args{-for}" if $args{-for};
+  # add FOR clause if needed
+  my $for = exists $args{-for} ? $args{-for} : $self->{select_implicitly_for};
+  $sql .= " FOR $for" if $for;
 
   if ($args{-want_details}) {
     return {sql             => $sql,
@@ -345,19 +346,10 @@ sub insert {
     my %args = validate(@_, \%params_for_insert);
     @old_API_args = @args{qw/-into -values/};
 
-    # if present, "-returning" may be a scalar, arrayref or hashref; the latter
-    # is interpreted as .. RETURNING ... INTO ...
-    if (my $returning = $args{-returning}) {
-      if (does($returning, 'HASH')) {
-        my @keys = sort keys %$returning
-          or croak "-returning => {} : the hash is empty";
-        push @old_API_args, {returning => \@keys};
-        $returning_into = [@{$returning}{@keys}];
-      }
-      else {
-        push @old_API_args, {returning => $returning};
-      }
-    }
+    # deal with -returning arg
+    ($returning_into, my $old_API_options) 
+      = $self->_compute_returning($args{-returning});
+    push @old_API_args, $old_API_options if $old_API_options;
   }
   else {
     @old_API_args = @_;
@@ -375,28 +367,73 @@ sub insert {
   return ($sql, @bind);
 }
 
+
 sub update {
   my $self = shift;
 
+  my $join_info;
   my @old_API_args;
+  my $returning_into;
   my %args;
   if (&_called_with_named_args) {
     %args = validate(@_, \%params_for_update);
+
+    # compute join info if the datasource is a join
+    $join_info = $self->_compute_join_info($args{-table});
+    $args{-table} = \($join_info->{sql}) if $join_info;
+
     @old_API_args = @args{qw/-table -set -where/};
+
+    # deal with -returning arg
+    ($returning_into, my $old_API_options) 
+      = $self->_compute_returning($args{-returning});
+    push @old_API_args, $old_API_options if $old_API_options;
   }
   else {
     @old_API_args = @_;
   }
 
-  # call clone of parent method
+  # call clone of parent method and merge with bind values from $join_info
   my ($sql, @bind) = $self->_overridden_update(@old_API_args);
+  unshift @bind, @{$join_info->{bind}} if $join_info;
 
   # maybe need to handle additional args
   $self->_handle_additional_args_for_update_delete(\%args, \$sql, \@bind);
 
+  # inject more stuff if using Oracle's "RETURNING ... INTO ..."
+  if ($returning_into) {
+    $sql .= ' INTO ' . join(", ", ("?") x @$returning_into);
+    push @bind, @$returning_into;
+  }
+
   return ($sql, @bind);
 }
 
+sub _compute_returning {
+  my ($self, $arg_returning) = @_; 
+
+  my ($returning_into, $old_API_options);
+
+  if ($arg_returning) {
+    # if present, "-returning" may be a scalar, arrayref or hashref; the latter
+    # is interpreted as .. RETURNING ... INTO ...
+
+
+    if (does $arg_returning, 'HASH') {
+      my @keys = sort keys %$arg_returning
+        or croak "-returning => {} : the hash is empty";
+
+      $old_API_options = {returning => \@keys};
+      $returning_into  = [@{$arg_returning}{@keys}];
+    }
+    else {
+      $old_API_options = {returning => $arg_returning};
+    }
+  }
+
+  return ($returning_into, $old_API_options);
+}
+
 
 sub _handle_additional_args_for_update_delete {
   my ($self, $args, $sql_ref, $bind_ref) = @_;
@@ -413,7 +450,27 @@ sub _handle_additional_args_for_update_delete {
   }
 }
 
+sub _order_by {
+  my ($self, $order) = @_;
 
+  # force scalar into an arrayref
+  $order = [$order] if not ref $order;
+
+  # restructure array data
+  if (does $order, 'ARRAY') {
+    my @clone = @$order;      # because we will modify items
+
+    # '-' and '+' prefixes are translated into {-desc/asc => } hashrefs
+    foreach my $item (@clone) {
+      next if !$item or ref $item;
+      $item =~ s/^-//  and $item = {-desc => $item} and next;
+      $item =~ s/^\+// and $item = {-asc  => $item};
+    }
+    $order = \@clone;
+  }
+
+  return $self->next::method($order);
+}
 
 sub delete {
   my $self = shift;
@@ -451,6 +508,14 @@ foreach my $attr (qw/table_alias column_alias limit_offset/) {
   };
 }
 
+# readonly accessor methods
+foreach my $key (qw/join_syntax  join_assoc_right
+                    max_members_IN multicols_sep  has_multicols_in_SQL/) {
+  no strict 'refs';
+  *{$key} = sub {shift->{$key}};
+}
+
+
 # invocation method for 'join'
 sub join {
   my $self = shift;
@@ -483,13 +548,13 @@ sub merge_conditions {
   my %merged;
 
   foreach my $cond (@_) {
-    if    (does($cond, 'HASH'))  {
+    if    (does $cond, 'HASH')  {
       foreach my $col (sort keys %$cond) {
         $merged{$col} = $merged{$col} ? [-and => $merged{$col}, $cond->{$col}]
                                       : $cond->{$col};
       }
     }
-    elsif (does($cond, 'ARRAY')) {
+    elsif (does $cond, 'ARRAY') {
       $merged{-nest} = $merged{-nest} ? {-and => [$merged{-nest}, $cond]}
                                       : $cond;
     }
@@ -507,12 +572,11 @@ sub bind_params {
   $sth->isa('DBI::st') or croak "sth argument is not a DBI statement handle";
   foreach my $i (0 .. $#bind) {
     my $val = $bind[$i];
-    my $ref = ref $val || '';
-    if ($ref eq 'SCALAR') {
+    if (does $val, 'SCALAR') {
       # a scalarref is interpreted as an INOUT parameter
       $sth->bind_param_inout($i+1, $val, $INOUT_MAX_LEN);
     }
-    elsif ($ref eq 'ARRAY' and
+    elsif (does $val, 'ARRAY' and
              my ($bind_meth, @args) = $self->is_bind_value_with_type($val)) {
       # either 'bind_param' or 'bind_param_inout', with 2 or 3 args
       $sth->$bind_meth($i+1, @args);
@@ -560,6 +624,19 @@ sub is_bind_value_with_type {
 # private utility methods for 'join'
 #----------------------------------------------------------------------
 
+sub _compute_join_info {
+  my ($self, $table_arg) = @_;
+
+  if (does($table_arg, 'ARRAY') && $table_arg->[0] eq '-join') {
+    my @join_args = @$table_arg;
+    shift @join_args;                # drop initial '-join'
+    return $self->join(@join_args);
+  }
+  else {
+    return;
+  }
+}
+
 sub _parse_table {
   my ($self, $table) = @_;
 
@@ -613,8 +690,8 @@ sub _parse_join_spec {
       or croak "can't parse join condition: $cond";
 
     # if operands are not qualified by table/alias name, add sprintf hooks
-    $left  = "%1\$s.$left"  unless $left  =~ /\./;
-    $right = "%2\$s.$right" unless $right =~ /\./ or $right eq $placeholder;
+    $left  = '%1$s.' . $left   unless $left  =~ /\./;
+    $right = '%2$s.' . $right  unless $right =~ /\./ or $right eq $placeholder;
 
     # add this pair into the list; right operand is either a bind value
     # or an identifier within the right table
@@ -637,15 +714,24 @@ sub _single_join {
   @_ = reverse @_ if $self->{join_assoc_right};
   my ($left, $join_spec, $right) = @_;
 
-  # compute the "ON" clause (assuming it contains '%1$s', '%2$s' for
-  # left/right tables)
+  # compute the "ON" clause
   my ($sql, @bind) = $self->where($join_spec->{condition});
   $sql =~ s/^\s*WHERE\s+//;
-  $sql = sprintf $sql, $left->{name}, $right->{name};
 
-  # assemble all elements
+  # syntax for assembling all elements
   my $syntax = $self->{join_syntax}{$join_spec->{operator}};
-  $sql = sprintf $syntax, $left->{sql}, $right->{sql}, $sql;
+
+  # $sql may _intentionally_ omit %.. parameters, so disable warnings
+  { no if $] ge '5.022000', warnings => 'redundant';
+
+    # substitute left/right tables names for '%1$s', '%2$s'
+    $sql = sprintf $sql, $left->{name}, $right->{name};
+
+    # build the final sql
+    $sql = sprintf $syntax, $left->{sql}, $right->{sql}, $sql;
+  }
+
+  # add left/right bind parameters (if any) into the list
   unshift @bind, @{$left->{bind}}, @{$right->{bind}};
 
   # build result and return
@@ -667,6 +753,22 @@ sub _single_join {
 sub _where_field_IN {
   my ($self, $k, $op, $vals) = @_;
 
+  # special algorithm if the key is multi-columns (contains a multicols_sep)
+  if ($self->{multicols_sep}) {
+    my @cols = split m[$self->{multicols_sep}], $k;
+    if (@cols > 1) {
+      if ($self->{has_multicols_in_SQL}) {
+        # DBMS accepts special SQL syntax for multicolumns
+        return $self->_multicols_IN_through_SQL(\@cols, $op, $vals);
+      }
+      else {
+        # DBMS doesn't accept special syntax, so we must use boolean logic
+        return $self->_multicols_IN_through_boolean(\@cols, $op, $vals);
+      }
+    }
+  }
+
+  # special algorithm if the number of values exceeds the allowed maximum
   my $max_members_IN = $self->{max_members_IN};
   if ($max_members_IN && does($vals, 'ARRAY')
                       &&  @$vals > $max_members_IN) {
@@ -682,12 +784,97 @@ sub _where_field_IN {
     $sql =~ s/\s*where\s*\((.*)\)/$1/i;
     return ($sql, @bind);
   }
-  else {
-    $vals = [@$vals] if blessed $vals; # because SQLA dies on blessed arrayrefs
-    return $self->next::method($k, $op, $vals);
-  }
+
+
+  # otherwise, call parent method
+  $vals = [@$vals] if blessed $vals; # because SQLA dies on blessed arrayrefs
+  return $self->next::method($k, $op, $vals);
 }
 
+
+sub _multicols_IN_through_SQL {
+  my ($self, $cols, $op, $vals) = @_;
+
+  # build initial sql
+  my $n_cols   = @$cols;
+  my $sql_cols = CORE::join(',', map {$self->_quote($_)} @$cols);
+  my $sql      = "($sql_cols) " . $self->_sqlcase($op);
+
+  # dispatch according to structure of $vals
+  return $self->_SWITCH_refkind($vals, {
+
+    ARRAYREF => sub {    # list of tuples
+      # deal with special case of empty list (like the parent class)
+      my $n_tuples = @$vals;
+      if (!$n_tuples) {
+        my $sql = ($op =~ /\bnot\b/i) ? $self->{sqltrue} : $self->{sqlfalse};
+        return ($sql);
+      }
+
+      # otherwise, build SQL and bind values for the list of tuples
+      my @bind;
+      foreach my $val (@$vals) {
+        does($val, 'ARRAY')
+          or $val = [split  m[$self->{multicols_sep}], $val];
+        @$val == $n_cols
+          or puke "op '$op' with multicols: tuple with improper number of cols";
+        push @bind, @$val;
+      }
+      my $single_tuple = "(" . CORE::join(',', (('?') x $n_cols)) . ")";
+
+      my $all_tuples   = CORE::join(', ', (($single_tuple) x $n_tuples));
+      $sql            .= " ($all_tuples)";
+      return ($sql, @bind);
+    },
+
+    SCALARREF => sub {   # literal SQL
+      $sql .= " ($$vals)";
+      return ($sql);
+    },
+
+    ARRAYREFREF => sub { # literal SQL with bind
+      my ($inner_sql, @bind) = @$$vals;
+      $sql .= " ($inner_sql)";
+      return ($sql, @bind);
+    },
+
+    FALLBACK => sub {
+      puke "op '$op' with multicols requires a list of tuples or literal SQL";
+    },
+
+   });
+}
+
+
+sub _multicols_IN_through_boolean {
+  my ($self, $cols, $op, $vals) = @_;
+
+  # can't handle anything else than a list of tuples
+  does($vals, 'ARRAY') && @$vals
+    or puke "op '$op' with multicols requires a non-empty list of tuples";
+
+  # assemble SQL
+  my $n_cols   = @$cols;
+  my $sql_cols = CORE::join(' AND ', map {$self->_quote($_) . " = ?"} @$cols);
+  my $sql      = "(" . CORE::join(' OR ', (("($sql_cols)") x @$vals)) . ")";
+  $sql         = "NOT $sql" if $op =~ /\bnot\b/i;
+
+  # assemble bind values
+  my @bind;
+  foreach my $val (@$vals) {
+    does($val, 'ARRAY')
+      or $val = [split  m[$self->{multicols_sep}], $val];
+    @$val == $n_cols
+      or puke "op '$op' with multicols: tuple with improper number of cols";
+    push @bind, @$val;
+  }
+
+  # return the whole thing
+  return ($sql, @bind);
+}
+
+
+
 #----------------------------------------------------------------------
 # override of parent's methods for decoding arrayrefs
 #----------------------------------------------------------------------
@@ -799,10 +986,11 @@ sub _overridden_update {
   my $table = $self->_table(shift);
   my $data  = shift || return;
   my $where = shift;
+  my $options = shift;
 
   # first build the 'SET' part of the sql statement
   my (@set, @all_bind);
-  SQL::Abstract::puke("Unsupported data type specified to \$sql->update")
+  puke "Unsupported data type specified to \$sql->update"
     unless ref $data eq 'HASH';
 
   for my $k (sort keys %$data) {
@@ -836,9 +1024,8 @@ sub _overridden_update {
       HASHREF => sub {
         my ($op, $arg, @rest) = %$v;
 
-        SQL::Abstract::puke(
-            'Operator calls in update must be in the form { -op => $arg }'
-           ) if (@rest or not $op =~ /^\-(.+)/);
+        puke 'Operator calls in update must be in the form { -op => $arg }'
+          if (@rest or not $op =~ /^\-(.+)/);
 
         local $self->{_nested_func_lhs} = $k;
         my ($sql, @bind) = $self->_where_unary_op ($1, $arg);
@@ -863,22 +1050,56 @@ sub _overridden_update {
     push @all_bind, @where_bind;
   }
 
+  if ($options->{returning}) {
+    my ($returning_sql, @returning_bind) = $self->_update_returning($options);
+    $sql .= $returning_sql;
+    push @all_bind, @returning_bind;
+  }
+
   return wantarray ? ($sql, @all_bind) : $sql;
 }
 
+
+
 #----------------------------------------------------------------------
 # method creations through closures
 #----------------------------------------------------------------------
 
-sub _make_AS_through_sprintf {
-  my ($self, $attribute) = @_;
-  my $syntax = $self->{$attribute};
-  $self->{$attribute} = sub {
+sub _make_sub_column_alias {
+  my ($self) = @_;
+  my $syntax = $self->{column_alias};
+  $self->{column_alias} = sub {
     my ($self, $name, $alias) = @_;
-    return $alias ? sprintf($syntax, $name, $alias) : $name;
+    return $name if !$alias;
+
+    # quote $name unless it is an SQL expression (then the user should quote it)
+    $name = $self->_quote($name) unless $name =~ /[()]/;
+
+    # assemble syntax
+    my $sql = sprintf $syntax, $name, $self->_quote($alias);
+
+    # return a string ref to avoid quoting by SQLA
+    return \$sql;
   };
 }
 
+
+sub _make_sub_table_alias {
+  my ($self) = @_;
+  my $syntax = $self->{table_alias};
+  $self->{table_alias} = sub {
+    my ($self, $name, $alias) = @_;
+    return $name if !$alias;
+
+    # assemble syntax
+    my $sql = sprintf $syntax, $self->_quote($name), $self->_quote($alias);
+
+    return $sql;
+  };
+}
+
+
+
 sub _choose_LIMIT_OFFSET_dialect {
   my $self = shift;
   my $dialect = $self->{limit_offset};
@@ -909,18 +1130,24 @@ SQL::Abstract::More - extension of SQL::Abstract with more constructs and more f
 
 Generates SQL from Perl data structures.  This is a subclass of
 L<SQL::Abstract>, fully compatible with the parent class, but with
-some additions :
+some improvements :
 
 =over
 
 =item *
 
-additional SQL constructs like C<-union>, C<-group_by>, C<join>, etc.
+methods take arguments as I<named parameters> instead of positional parameters.
+This is more flexible for identifying and assembling various SQL clauses,
+like C<-where>, C<-order_by>, C<-group_by>, etc.
 
 =item *
 
-methods take arguments as named parameters instead of positional parameters,
-so that various SQL fragments are more easily identified
+additional SQL constructs like C<-union>, C<-group_by>, C<join>, etc.
+are supported
+
+=item *
+
+C<WHERE .. IN> clauses can range over multiple columns (tuples)
 
 =item *
 
@@ -928,13 +1155,20 @@ values passed to C<select>, C<insert> or C<update> can directly incorporate
 information about datatypes, in the form of arrayrefs of shape
 C<< [{dbd_attrs => \%type}, $value] >>
 
-=back
+=item *
 
+several I<SQL dialects> can adapt the generated SQL to various DBMS vendors
+
+=back
 
 This module was designed for the specific needs of
 L<DBIx::DataModel>, but is published as a standalone distribution,
 because it may possibly be useful for other needs.
 
+Unfortunately, this module cannot be used with L<DBIx::Class>, because
+C<DBIx::Class> creates its own instance of C<SQL::Abstract>
+and has no API to let the client instantiate from any other class.
+
 =head1 SYNOPSIS
 
   my $sqla = SQL::Abstract::More->new();
@@ -976,6 +1210,16 @@ because it may possibly be useful for other needs.
   $sqla->bind_params($sth, @bind);
   $sth->execute;
 
+  # ex5: multicolumns-in
+  $sqla = SQL::Abstract::More->new(
+    multicols_sep        => '/',
+    has_multicols_in_SQL => 1,
+  );
+  ($sql, @bind) = $sqla->select(
+   -from     => 'Foo',
+   -where    => {"foo/bar/buz" => {-in => ['1/a/X', '2/b/Y', '3/c/Z']}},
+  );
+
   # merging several criteria
   my $merged = $sqla->merge_conditions($cond_A, $cond_B, ...);
   ($sql, @bind) = $sqla->select(..., -where => $merged, ..);
@@ -1066,6 +1310,76 @@ C<-not_in> operator).
   #            AND (bar NOT IN (?,?,?) AND bar NOT IN (?, ?)) )
 
 
+=item multicols_sep
+
+A string or compiled regular expression used as a separator for
+"multicolumns". This separator can then be used on the left-hand side
+and right-hand side of an C<IN> operator, like this :
+
+  my $sqla = SQL::Abstract::More->new(multicols_sep => '/');
+  ($sql, @bind) = $sqla->select(
+   -from     => 'Foo',
+   -where    => {"x/y/z" => {-in => ['1/A/foo', '2/B/bar']}},
+  );
+
+Alternatively, tuple values on the right-hand side can also be given
+as arrayrefs instead of plain scalars with separators :
+
+   -where    => {"x/y/z" => {-in => [[1, 'A', 'foo'], [2, 'B', 'bar']]}},
+
+but the left-hand side must stay a plain scalar because an array reference
+wouldn't be a proper key for a Perl hash; in addition, the presence
+of the separator in the string is necessary to trigger the special
+algorithm for multicolumns.
+
+The generated SQL depends on the boolean flag C<has_multicols_in_SQL>,
+as explained in the next paragraph.
+
+=item has_multicols_in_SQL
+
+A boolean flag that controls which kind of SQL will be generated for
+multicolumns. If the flag is B<true>, this means that the underlying DBMS
+supports multicolumns in SQL, so we just generate tuple expressions.
+In the example from the previous paragraph, the SQL and bind values
+would be :
+
+   # $sql  : "WHERE (x, y, z) IN ((?, ?, ?), (?, ?, ?))"
+   # @bind : [ qw/1 A foo 2 B bar/ ]
+
+It is also possible to use a subquery, like this :
+
+  ($sql, @bind) = $sqla->select(
+   -from     => 'Foo',
+   -where    => {"x/y/z" => {-in => \[ 'SELECT (a, b, c) FROM Bar '
+                                       . 'WHERE a > ?', 99]}},
+  );
+  # $sql  : "WHERE (x, y, z) IN (SELECT (a, b, c) FROM Bar WHERE a > ?)"
+  # @bind : [ 99 ]
+
+If the flag is B<false>, the condition on tuples will be
+automatically converted using boolean logic :
+
+   # $sql  : "WHERE (   (x = ? AND y = ? AND z = ?) 
+                     OR (x = ? AND y = ? AND z = ?))"
+   # @bind : [ qw/1 A foo 2 B bar/ ]
+
+If the flag is false, subqueries are not allowed.
+
+
+=item select_implicitly_for
+
+A value that will be automatically added as a C<-for> clause in
+calls to L</select>. This default clause can always be overridden
+by an explicit C<-for> in a given select :
+
+  my $sqla = SQL::Abstract->new(-select_implicitly_for => 'READ ONLY');
+  ($sql, @bind) = $sqla->select(-from => 'Foo');
+    # SELECT * FROM FOO FOR READ ONLY
+  ($sql, @bind) = $sqla->select(-from => 'Foo', -for => 'UPDATE');
+    # SELECT * FROM FOO FOR UPDATE
+  ($sql, @bind) = $sqla->select(-from => 'Foo', -for => undef);
+    # SELECT * FROM FOO
+
 =item sql_dialect
 
 This is actually a "meta-argument" : it injects a collection
@@ -1165,6 +1479,21 @@ using perl C<< qw/.../ >> operator for columns, as in
   -columns => [ qw/table1.longColumn|t1lc table2.longColumn|t2lc/ ]
 
 SQL column aliasing is then generated through the L</column_alias> method.
+If L<SQL::Abstract/quote_char> is defined, aliased columns will be quoted,
+unless they contain parentheses, in which case they are considered as
+SQL expressions for which the user should handle the quoting himself.
+For example if C<quote_char> is "`", 
+
+  -columns => [ qw/foo.bar|fb length(buz)|lbuz/ ]
+
+will produce
+
+  SELECT `foo`.`bar` AS fb, length(buz) AS lbuz
+
+and not
+
+  SELECT `foo`.`bar` AS fb, length(`buz`) AS lbuz
+
 
 Initial items in C<< @columns >> that start with a minus sign
 are shifted from the array, i.e. they are not considered as column
@@ -1237,7 +1566,7 @@ specified either by a plain string or by an array of strings.
 adds a C<HAVING> clause in the SQL statement (only makes
 sense together with a C<GROUP BY> clause).
 This is like a C<-where> clause, except that the criteria
-are applied after grouping has occured.
+are applied after grouping has occurred.
 
 
 =item C<< -order_by => \@order >>
@@ -1377,11 +1706,12 @@ present module is there for help. Example:
 
   # named parameters, handled in this class
   ($sql, @bind) = $sqla->update(
-    -table    => $table,
-    -set      => {col => $val, ...},
-    -where    => \%conditions,
-    -order_by => \@order,
-    -limit    => $limit,
+    -table     => $table,
+    -set       => {col => $val, ...},
+    -where     => \%conditions,
+    -order_by  => \@order,
+    -limit     => $limit,
+    -returning => $return_structure,
   );
 
 This works in the same spirit as the L</insert> method above.
@@ -1392,6 +1722,8 @@ they improve the readability of the client's code.
 Few DBMS would support parameters C<-order_by> and C<-limit>, but
 MySQL does -- see L<http://dev.mysql.com/doc/refman/5.6/en/update.html>.
 
+Optional parameter C<-returning> works like for the L</insert> method.
+
 
 =head2 delete
 
@@ -1503,6 +1835,7 @@ SQL JOIN clauses :
    '=>' => '%s LEFT OUTER JOIN %s ON %s',
   '<='  => '%s RIGHT JOIN %s ON %s',
   '=='  => '%s NATURAL JOIN %s',
+  '>=<' => '%s FULL OUTER JOIN %s ON %s',
 
 This operator table can be overridden through
 the C<join_syntax> parameter of the L</new> method.
@@ -1707,50 +2040,15 @@ L</bind_params> from the present module to perform the appropriate
 bindings before executing the statement.
 
 
-=head1 TODO
-
-Future versions may include some of these features :
-
-=over
-
-=item *
-
-support for C<WITH> initial clauses, and C<WITH RECURSIVE>.
-
-=item *
-
-support for Oracle-specific syntax for recursive queries
-(START_WITH, PRIOR, CONNECT_BY NOCYCLE, CONNECT SIBLINGS, etc.)
-
-=item *
-
-support for INSERT variants
-
-    INSERT .. DEFAULT VALUES
-    INSERT .. VALUES(), VALUES()
-
-=item *
-
-support for MySQL C<LOCK_IN_SHARE_MODE>
-
-=item *
-
-new constructor option 
-
-  ->new(..., select_implicitly_for => $string, ...)
-
-This would provide a default values for the C<-for> parameter.
-
-=back
-
 =head1 AUTHOR
 
-Laurent Dami, C<< <laurent.dami at justice.ge.ch> >>
+Laurent Dami, C<< <laurent dot dami at cpan dot org> >>
 
 =head1 BUGS
 
 Please report any bugs or feature requests to C<bug-sql-abstract-more at rt.cpan.org>, or through
-the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=SQL-Abstract-More>.  I will be notified, and then you'll
+the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=SQL-Abstract-More>. 
+I will be notified, and then you'll
 automatically be notified of progress on your bug as I make changes.
 
 
@@ -1786,7 +2084,7 @@ L<https://metacpan.org/module/SQL::Abstract::More>
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2011, 2012 Laurent Dami.
+Copyright 2011-2017 Laurent Dami.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
